

class Environment:
    def __init__(self):
        self.parent: Environment | None = None
        self.store = {}  # id-val, id-mixin

    def get(self, name: str):
        obj = None
        if self.parent:
            obj = self.store.get(name)

        obj0 = self.store.get(name)

        return obj0 if obj0 else obj

    def set(self, name: str, val):
        self.store[name] = val
        return val




from pqss.env import Environment
from pqss.lex import Token
from .ast import (
    Statement,
    Expression
)
from .identifier import Identifier


class VarStatement(Statement):
    def eval(self, environment: Environment):
        environment.set(self.name.token.literal,
                        self.value.eval(environment))

    def __init__(self):
        self.token: Token | None = None  # TODO May be not needed
        self.name: Identifier | None = None
        self.value: Expression | None = None

    def stmt_node(self):
        pass



import logging
import re

from .token import Token, TokenType, lookup_keyword, is_color, is_property
from .utils import is_digit, is_letter, is_white_space
from .exceptions import TokenUnKnownException
from .constants import EOF

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())


class Lexer:
    """Lexer of PQSS"""
    def __init__(self, src_code: str):
        """
        :param src_code:  PQSS code
        """
        self._src_code: str = src_code
        self._cur_pos: int = 0
        self._peek_pos: int = 0
        self._cur_char = None
        self._peek_char = None

        self.read_char()

    def read_char(self) -> str:
        """read a char and preview one char"""
        if self._peek_pos >= len(self._src_code):  # 到达文本末尾，全部读作EOF
            self._peek_char = EOF
            self._cur_char = EOF
        else:
            self._cur_char = self._src_code[self._peek_pos]
            self._cur_pos = self._peek_pos
            self._peek_pos += 1
            if not self._peek_pos >= len(self._src_code):  # 到达文本末尾，字符不需要再改变了， 一律EOF
                self._peek_char = self._src_code[self._peek_pos]

        return self._cur_char

    def next_token(self) -> Token:
        """parse a lexeme to Token"""
        self._skip_blank_and_comment()

        lexeme = self._cur_char
        tok = None

        if lexeme == EOF:
            tok = Token(TokenType.EOF, lexeme)

        elif lexeme == ':':
            tok = Token(TokenType.ASSIGN, lexeme)
        elif lexeme == '(':
            tok = Token(TokenType.LEFT_PAREN, lexeme)
        elif lexeme == ')':
            tok = Token(TokenType.RIGHT_PAREN, lexeme)
        elif lexeme == '{':
            tok = Token(TokenType.LEFT_BRACE, lexeme)
        elif lexeme == '}':
            tok = Token(TokenType.RIGHT_BRACE, lexeme)
        elif lexeme == ';':
            tok = Token(TokenType.SEMICOLON, lexeme)
        elif lexeme == ',':
            tok = Token(TokenType.COMMA, lexeme)

        elif lexeme == '+':
            tok = Token(TokenType.PLUS, lexeme)
        elif lexeme == '-':
            tok = Token(TokenType.SUB, lexeme)
        elif lexeme == '*':
            tok = Token(TokenType.MUL, lexeme)
        elif lexeme == '/':
            tok = Token(TokenType.DIV, lexeme)
        elif lexeme == '>':
            tok = Token(TokenType.GT, lexeme)
        elif lexeme == '<':
            tok = Token(TokenType.LT, lexeme)
        elif lexeme == '=':
            lexeme = '=='
            self.read_char()
            tok = Token(TokenType.EQ, lexeme)

        elif lexeme == '"' or lexeme == "'":
            lexeme = self.read_string()
            tok = Token(TokenType.STRING, lexeme)

        elif lexeme == '&':
            tok = Token(TokenType.PARENT_REFERENCE, lexeme)
        elif lexeme == '$':
            lexeme = self.read_identifier()
            tok = Token(TokenType.IDENTIFIER, lexeme)
        elif lexeme == '!' or lexeme == '@':
            if self._peek_char == '=':
                lexeme = '!='
                self.read_char()
                tok = Token(TokenType.EQ, lexeme)
            else:
                tok = self.read_keyword()
        elif lexeme == '.':
            lexeme = self.read_ID_selector()
            tok = Token(TokenType.TYPE_SELECTOR, lexeme)
        elif lexeme == '*':
            tok = Token(TokenType.GENERAL_SELECTOR, lexeme)
        elif lexeme == '#':
            if self.is_value():
                lexeme = self.read_word()
                tok = Token(TokenType.COLOR, lexeme)
            else:
                lexeme = self.read_ID_selector()
                tok = Token(TokenType.ID_SELECTOR, lexeme)
        elif lexeme == '>':
            tok = Token(TokenType.CHILDREN_SELECTOR, lexeme)

        elif is_digit(lexeme):
            lexeme = self.read_number()
            tok = Token(TokenType.NUMBER, lexeme)

        elif lexeme == '%':
            raise NotImplementedError()

        elif is_letter(lexeme):
            if self.is_color():
                lexeme = self.read_word()
                tok = Token(TokenType.COLOR, lexeme)
            elif self.is_property():
                lexeme = self.read_property()
                tok = Token(TokenType.PROPERTY, lexeme)
            elif self.is_keyword():
                tok = self.read_keyword()
            elif self.is_builtin():
                lexeme = self.read_word()
                tok = Token(TokenType.BUILTIN, lexeme)
            elif self.is_mixin_name():
                lexeme = self.read_identifier()
                tok = Token(TokenType.IDENTIFIER, lexeme)
            else:
                tok = self.read_selector()
        else:
            raise TokenUnKnownException(f'Token {lexeme} does unknown!!!')

        self.read_char()
        return tok

    def read_identifier(self):
        """read a valid identifier"""
        pos = self._cur_pos
        while is_letter(self._peek_char):
            self.read_char()
        return self._src_code[pos:self._peek_pos]

    def read_number(self):
        """read a valid number"""
        pos = self._cur_pos
        while is_digit(self._peek_char):  # int
            self.read_char()

        if self._peek_char == '.':  # float
            while is_digit(self._peek_char):
                self.read_char()
        return self._src_code[pos:self._peek_pos]

    def is_keyword(self) -> bool:
        """check the next word is a valid token for PQSS"""
        pos = self._cur_pos

        while is_letter(self._peek_char):
            self.read_char()
        lexeme = self._src_code[pos: self._cur_pos]

        # restore
        self._cur_pos = pos
        self._cur_char = self._src_code[self._cur_pos]
        self._peek_pos = pos + 1
        self._peek_char = self._src_code[self._peek_pos]

        if lookup_keyword(lexeme):
            return True
        return False

    def read_keyword(self):
        """read a keyword to Token"""
        pos = self._cur_pos

        while is_letter(self._peek_char):
            self.read_char()
        lexeme = self._src_code[pos:self._peek_pos]

        token_type = lookup_keyword(lexeme)
        if token_type is None:
            raise TokenUnKnownException(f'Token {lexeme} does not a valid keyword!!!')
        return Token(token_type, lexeme)

    def read_selector(self):
        """read the non-prefix selector"""
        pos = self._cur_pos
        token_type = None
        while is_letter(self.read_char()):
            pass

        # 属性选择器
        if self._cur_char == '[':
            while self._cur_char != ']':
                self.read_char()
            self.read_char()  # 读掉 ]
            token_type = TokenType.PROPERTY_SELECTOR
        elif self._cur_char == ':':
            if self.read_char() == ':':
                token_type = TokenType.SUBWIDGET_SELECTOR
            else:
                token_type = TokenType.PRODO_SELECTOR
            while is_letter(self.read_char()):
                pass
        else:
            token_type = TokenType.CLASS_SELECTOR
        return Token(token_type, self._src_code[pos:self._cur_pos])

    def read_ID_selector(self):
        """read the selector begin with #"""
        pos = self._cur_pos
        self.read_char()
        while is_letter(self._peek_char):
            self.read_char()
        return self._src_code[pos:self._cur_pos + 1]

    def is_property(self) -> bool:
        """check if the next word is property of QSS"""
        pos = self._cur_pos
        while is_letter(self.read_char()):
            pass
        lexeme = self._src_code[pos: self._cur_pos]
        self._cur_pos = pos
        self._cur_char = self._src_code[self._cur_pos]
        self._peek_pos = pos + 1
        self._peek_char = self._src_code[self._peek_pos]

        if is_property(lexeme):
            return True
        return False

    def read_property(self) -> str:
        """read the next properties and return the lexeme"""
        pos = self._cur_pos
        while is_letter(self._peek_char):
            self.read_char()
        lexeme = self._src_code[pos: self._peek_pos]
        if is_property(lexeme):
            return lexeme
        self._cur_pos = pos
        self._cur_char = self._src_code[pos]

    def is_mixin_name(self):
        """check if the next identifier belongs to a mixin"""
        i = 0
        while is_letter(self._src_code[self._cur_pos + i]):
            i += 1
        while is_white_space(self._src_code[self._cur_pos + i]):
            i += 1
        ch = self._src_code[self._cur_pos + i]
        return ch == '('

    def _skip_blank_char(self):
        """skip all the next black chars"""
        while is_white_space(self._cur_char):
            self.read_char()

    def _skip_comment(self):
        """skip comments, // or /**/"""
        if self._cur_char == '/':
            if self._peek_char == '/':
                while self._cur_char != '\n':
                    self.read_char()
                self.read_char()  # skip \n
            elif self._peek_char == '*':
                stack = 1
                while stack != 0:
                    self.read_char()
                    if self._cur_char == '/' and self._peek_char == '*':
                        stack += 1
                    elif self._cur_char == '*' and self._peek_char == '/':
                        stack -= 1
                # skip */
                self.read_char()
                self.read_char()

    def _skip_blank_and_comment(self):
        self._skip_blank_char()
        self._skip_comment()
        self._skip_blank_char()

    def is_end(self):
        return self._cur_char == EOF

    def read_string(self):
        pos= self._peek_pos
        quote = self._cur_char

        while self._peek_char != quote:
            self.read_char()

        return self._src_code[pos: self._peek_pos]

    def is_color(self):
        pos = self._cur_pos

        while is_letter(self._peek_char):
            self.read_char()

        lexeme = self._src_code[pos: self._peek_pos]
        self._cur_pos = pos
        self._cur_char = self._src_code[self._cur_pos]
        self._peek_pos = pos + 1
        self._peek_char = self._src_code[self._peek_pos]

        return is_color(lexeme)

    def read_word(self):
        pos = self._cur_pos
        while is_letter(self._peek_char):
            self.read_char()
        lexeme = self._src_code[pos: self._peek_pos]

        return lexeme

    def is_value(self):
        i = 0
        while not is_white_space(self._src_code[self._cur_pos + i]) and self._src_code[self._cur_pos + i] != ';':
            i += 1
        ch = self._src_code[self._cur_pos + i]
        return ch == ';'

    def is_builtin(self):
        pos = self._cur_pos

        while is_letter(self._peek_char):
            self.read_char()

        lexeme = self._src_code[pos: self._peek_pos]
        self._cur_pos = pos
        self._cur_char = self._src_code[self._cur_pos]
        self._peek_pos = pos + 1
        self._peek_char = self._src_code[self._peek_pos]

        return lexeme in ['rgb', 'rgba']





import abc

from pqss.lex import TokenType, Token
from pqss.env import *


class Node:
    @abc.abstractmethod
    def eval(self, environment: Environment):
        pass


class Statement(Node):
    @abc.abstractmethod
    def stmt_node(self):
        pass


class Expression(Node):
    @abc.abstractmethod
    def expr_node(self):
        pass


class StyleSheet(Node):
    def __init__(self):
        self.statements: list[Statement] = []

    def eval(self, environment: Environment):
        qss = ''
        for stmt in self.statements:
            result = stmt.eval(environment)
            if result is not None:
                qss = qss + str(result)

        return qss


class ExpressionStatement(Statement):
    def __init__(self):
        self.expr: Expression | None = None

    def eval(self, environment: Environment):
        return self.expr.eval(environment)

    def stmt_node(self):
        pass


class PrefixExpression(Expression):
    def eval(self, environment: Environment):
        right = self.right.eval(environment)
        if self.token.token_type == TokenType.SUB:
            return -right

    def __init__(self, token: Token | None, operator: str | None = None, right: Expression | None = None):
        self.token = token
        self.operator = operator
        self.right = right

    def expr_node(self):
        pass


class InfixExpression(Expression):
    def eval(self, environment: Environment):
        left = self.left.eval(environment)
        right = self.right.eval(environment)
        if self.token.token_type == TokenType.PLUS:
            return left + right
        if self.token.token_type == TokenType.SUB:
            return left - right
        if self.token.token_type == TokenType.MUL:
            return left * right
        else:
            # self.token.token_type == TokenType.DIV:
            return left / right

    def __init__(self, token: Token, operator: str, left: Expression = None, right: Expression = None):
        self.token = token
        self.left = left
        self.operator = operator
        self.right = right

    def expr_node(self):
        pass

    def token_literal(self) -> str:
        pass


class BlockStatement(Statement):
    def eval(self, environment: Environment):
        res = '{'
        for stmt in self.statements:
            res += str(stmt.eval(environment))
        res += '}'
        return res

    def __init__(self, token: Token | None):
        self.token = token
        self.statements: list[Statement] = []

    def stmt_node(self):
        pass


class IfStatement(Statement):
    def eval(self, environment: Environment):
        pass

    def __init__(self, token: Token, expr: Expression | None = None,
                 consequence: BlockStatement | None = None,
                 alternative: BlockStatement | None = None):
        self.token = token
        self.condition = expr
        self.consequence = consequence
        self.alternative = alternative

    def stmt_node(self):
        pass






from pqss.env import Environment
from pqss.lex import Token
from .ast import (
    Statement,
    Expression,
    BlockStatement
)


class Include(Statement):
    def eval(self, environment: Environment):
        mixin = environment.get(self.mixin_name)
        if not mixin:
            raise "Mixin not exists!!!"
        params = mixin.params
        body = mixin.body

        return self.do_eval(params, body, environment)

    def do_eval(self, params, body, environment: Environment):
        bloc_env = Environment()
        bloc_env.parent = environment
        idx = 0
        while idx < len(params):
            key = params[idx].token.literal
            val = self.args[idx].eval(bloc_env)
            bloc_env.set(key, val)

            idx += 1

        s = body.eval(bloc_env)
        res = s[1:-1] if len(s) > 2 else ''
        return res

    def __init__(self, token: Token, mixin_name: str | None = None, args: list[Expression] | None = None):
        self.token = token
        self.mixin_name = mixin_name
        self.args = args

    def stmt_node(self):
        pass




from pqss.env import Environment
from pqss.lex import Token
from .ast import (
    Statement,
    BlockStatement
)
from .identifier import Identifier


class Mixin(Statement):
    def eval(self, environment: Environment):
        environment.set(self.name, self)

    def stmt_node(self):
        pass

    def __init__(self, token: Token, name: str | None = None):
        self.token = token
        self.name: str = name
        self.params: list[Identifier] = []
        self.body: BlockStatement | None = None



from enum import Enum, auto


class TokenType(Enum):
    BUILTIN = auto()
    ILLEGAL = auto()
    """Unknown Token"""
    EOF = auto()
    """End of File"""

    ASSIGN = auto()
    """Assign, $var: val"""
    PLUS = auto()
    """ + """
    SUB = auto()
    """ a-b, or -a """
    MUL = auto()
    """ * """
    DIV = auto()
    """ / """
    EQ = auto()
    """ == """
    NOT_EQ = auto()
    """ != """
    LT = auto()
    """ < """
    GT = auto()
    """ > """

    TRUE = auto()
    """true"""
    FALSE = auto()
    """false"""

    IF = auto()
    """@if"""
    ELSE = auto()
    """@else"""

    IDENTIFIER = auto()
    """ name of variable"""
    SEMICOLON = auto()
    """ ; """
    COMMA = auto()
    """ , """
    LEFT_PAREN = auto()
    """("""
    RIGHT_PAREN = auto()
    """)"""
    LEFT_BRACE = auto()
    """{"""
    RIGHT_BRACE = auto()
    """}"""

    NUMBER = auto()
    """Number"""
    STRING = auto()
    """String"""

    GENERAL_SELECTOR = auto()  # *
    TYPE_SELECTOR = auto()  # .MyButton
    CLASS_SELECTOR = auto()  # QPushButton
    ID_SELECTOR = auto()  # #btn
    PROPERTY_SELECTOR = auto()  # QPushButton[name='abc']
    CHILDREN_SELECTOR = auto()  # >
    GROUP_SELECTOR = auto()  # #a, #b
    PARENT_REFERENCE = auto()
    # POSTERITY_SELECTOR = auto()
    PRODO_SELECTOR = auto()  # 伪类选择器 :hover
    SUBWIDGET_SELECTOR = auto()  # 子组件选择器 ::indicator
    COLOR = auto()

    # keywords
    IMPORT = auto()  # @import
    DEFAULT = auto()
    EXTEND = auto()
    MIXIN = auto()
    INCLUDE = auto()

    PROPERTY = auto()


EOF = 0

keywords = {
    '@import': TokenType.IMPORT,
    '!default': TokenType.DEFAULT,
    '@extend': TokenType.EXTEND,
    '@mixin': TokenType.MIXIN,
    '@include': TokenType.INCLUDE,
    '@if': TokenType.IF,
    '@else': TokenType.ELSE,
    'true': TokenType.TRUE,
    'false': TokenType.FALSE
}
""" Keywords of PQSS"""

colors = ['red',
          'blue',
          'white',
          'yellow',
          'green',
          'gray',
          'black',
          'orange',
          'grown']
"""The string for color """

properties = ['alternate-background-color',
              'background',
              'background-color',
              'background-image',
              'background-repeat',
              'background-position',
              'background-attachment',
              'background-clip',
              'background-origin',
              'border',
              'border-top',
              'border-right',
              'border-bottom',
              'border-left',
              'border-color',
              'border-top-color',
              'border-right-color',
              'border-bottom-color',
              'border-left-color',
              'border-image',
              'border-radius',
              'border-top-left-radius',
              'border-top-right-radius',
              'border-bottom-right-radius',
              'border-bottom-left-radius',
              'border-style',
              'border-top-style',
              'border-right-style',
              'border-bottom-style',
              'border-left-style',
              'border-width',
              'border-top-width',
              'border-right-width',
              'border-bottom-width',
              'border-left-width',
              'bottom',
              'button-layout',
              'color',
              'dialogbuttonbox-buttons-have-icons',
              'font',
              'font-family',
              'font-family',
              'font-size',
              'font-style',
              'font-weight',
              'height',
              'icon-size',
              'image',
              'image-position',
              'lineedit-password-character',
              'margin',
              'margin-top',
              'margin-right',
              'margin-bottom',
              'margin-left',
              'max-height',
              'max-width',
              'messagebox-text-interaction-flags',
              'min-height',
              'min-width',
              'opacity',
              'padding',
              'padding-top',
              'padding-right',
              'padding-bottom',
              'padding-left',
              'paint-alternating-row-color-for-empty-area',
              'position',
              'right',
              'selection-background-color',
              'selection-color',
              'show-decoration-selected',
              'spacing',
              'subcontrol-origin',
              'text-align',
              'text-decoration',
              'top',
              'width']
"""Qss properties"""



from .ast import (
    Expression,
)
from pqss.env import Environment
from pqss.lex import Token


class Builtin(Expression):
    def eval(self, environment: Environment):
        res = self.token.literal
        res += '('
        for arg in self.args:
            res += str(arg.eval(environment)) + ','
        res = res[:-1] + ');'
        return res

    def __init__(self, token: Token, args: list[Expression] | None = None):
        self.token = token
        self.args = args

    def token_literal(self) -> str:
        return self.token.literal

    def expr_node(self):
        pass



from .constants import (
    colors,
    properties,
    keywords,
    TokenType
)
"""
    
"""


def is_property(lexeme: str):
    """check if a literal is a QSS property"""
    return lexeme in properties


def is_color(lexeme: str):
    """check if a literal is a color string"""
    return lexeme in colors


def lookup_keyword(lexeme):
    """
        :param lexeme literal string
        :return if the lexeme if a keyword, return the TokenType, or return None
    """
    return keywords.get(lexeme)


class Token:
    """Token Entry"""
    def __init__(self, token_type: TokenType | None, lexeme):
        self.token_type = token_type
        self.literal = lexeme

    def __eq__(self, other):
        return (type(other) is Token
                and other.token_type == self.token_type
                and other.literal == self.literal)




from pqss.env import Environment
from pqss.lex import Token
from .ast import (
    Statement,
    Expression,
)
from .selector import Selector
from .rule import Rule
from .include import Include


class Ruleset(Statement):
    def eval(self, environment: Environment):
        res = self._eval_selectors(environment)
        res += '{'
        for rule in self.rules:
            res += rule.eval(environment)
        for inc in self.includes:
            res += inc.eval(environment)
        res += '}'

        res += self._eval_child_rulesets(environment)

        return res

    def _eval_selectors(self, environment):
        res = ''
        for selector in self.selectors:
            res += ' ' + selector.eval(environment)
        return res

    def _eval_child_rulesets(self, environment):
        res = ''

        for selector in self.selectors[:-2]:
            res += selector.eval(environment)

        for ruleset in self.child_rulesets:
            val = ruleset.eval(environment)

            if val.find('&') != -1:
                val = val.replace('&', self.selectors[-1].eval(environment))
                res += ' ' + val
            else:
                res += self.selectors[-1].eval(environment) + val

        return res

    def __init__(self):
        self.selectors: list[Selector] | None = []
        self.rules: list[Rule] = []
        self.child_rulesets: list[Ruleset] = []
        self.includes: list[Include] = []


    def stmt_node(self):
        pass




def is_letter(ch):
    """check if a char is valid to consist identifier"""
    if type(ch) is not str:
        return False
    return ('a' <= ch <= 'z'
            or 'A' <= ch <= 'Z'
            or '0' <= ch <= '9'
            or ch in ['-', '_'])


def is_digit(ch):
    """check if a char is a number"""
    if type(ch) is not str:
        return False
    return '0' <= ch <= '9'


def is_white_space(ch):
    """check is a char is a blank char"""
    return (ch == ' '
            or ch == '\t'
            or ch == '\n'
            or ch == '\r')




from pqss.env import Environment
from pqss.lex import Token
from .ast import (
    Statement,
)


class Selector(Statement):
    def eval(self, environment: Environment):
        return self.token.literal

    def __init__(self, token: Token | None):
        self.token = token

    def stmt_node(self):
        pass




from pqss.env import Environment
from pqss.lex import Token
from .ast import (
    Statement,
    Expression,
)


class Rule(Statement):
    def eval(self, environment: Environment):
        return f'{self.property.literal}:{self.value.eval(environment)};'

    def __init__(self):
        self.property: Token | None = None
        self.value: Expression | None = None

    def stmt_node(self):
        pass





from pqss.env import Environment
from pqss.lex import Token
from .ast import (
    Expression
)


class Boolean(Expression):
    def eval(self, environment: Environment):
        return self.value

    def expr_node(self):
        pass

    def __init__(self, token: Token, value: bool):
        self.token = token
        self.value = value



from .ast import (
    Expression,
)
from pqss.env import Environment
from pqss.lex import Token


class Color(Expression):
    def eval(self, environment: Environment):
        return self.token.literal

    def __init__(self, token: Token):
        self.token: Token | None = token

    def expr_node(self):
        pass



from distutils.core import setup

setup(
    name='PQSS',
    packages=['pqss', 'pqss.env', 'pqss.lex', 'pqss.parse', 'pqss.parse.ast'],
    author='lyt0628',
    author_email='lyt.0628@qq.com',
    url='http://lyt0628.icu/docs/pqss',
    download_url='http://pypi.org//pqss',
    description='PQSS is a dynamic language for qss, like scss for css.'
)


from .ast import (
    Expression,
)
from pqss.env import Environment
from pqss.lex import Token


class Identifier(Expression):
    def eval(self, environment: Environment):
        val = environment.get(self.token.literal)
        if val:
            return val
        else:
            # environment.set(self.token.literal, self.value)
            return self.value

    def __init__(self, token: Token, value):
        self.token: Token | None = token
        self.value: str = ''

    def token_literal(self) -> str:
        return self.token.literal

    def expr_node(self):
        pass



import os.path

from pqss.lex import *
from pqss.env import *
from pqss.parse import *


def parse(source: str) -> str:
    if os.path.isfile(source):
        source = read_file(source)

    compiler = Parser(Lexer(source))
    style_sheet = compiler.parse_program()
    qss = style_sheet.eval(Environment())
    return qss


def read_file(p: str):
    res = ''
    with open(p, 'r') as f:
        res = f.read()
    return res




from pqss.lex import *
from pqss.parse.ast import *


# TODO
class Priority:
    LOWEST = 1
    EQUALS = 2
    LESS_GREATER = 3
    SUM = 4
    PRODUCT = 5
    PREFIX = 6
    CALL = 7


precedences = {
    TokenType.EQ: Priority.EQUALS,
    TokenType.NOT_EQ: Priority.EQUALS,
    TokenType.LT: Priority.LESS_GREATER,
    TokenType.GT: Priority.LESS_GREATER,
    TokenType.PLUS: Priority.SUM,
    TokenType.SUB: Priority.SUM,
    TokenType.MUL: Priority.PRODUCT,
    TokenType.DIV: Priority.PRODUCT
}


class Parser:
    def __init__(self, lex: Lexer):
        self.code_path: str | None = None

        self.sqss: StyleSheet | None = None
        self.lexer: Lexer = lex
        self.cur_token: Token | None = None
        self.peek_token: Token | None = None

        self.next_token()
        self.next_token()

        self.prefix_parse_fns = {}
        self.infix_parse_fns = {}

        self.register_prefix_fn(TokenType.IDENTIFIER, self.parse_identifier)
        self.register_prefix_fn(TokenType.COLOR, self.parse_color)
        self.register_prefix_fn(TokenType.BUILTIN, self.parse_builtin)
        self.register_prefix_fn(TokenType.NUMBER, self.parse_integer_literal)
        self.register_prefix_fn(TokenType.SUB, self.parse_prefix_expr)
        self.register_prefix_fn(TokenType.TRUE, self.parse_boolean)
        self.register_prefix_fn(TokenType.FALSE, self.parse_boolean)
        self.register_prefix_fn(TokenType.LEFT_PAREN, self.parse_grouped_expr)

        self.register_infix_fn(TokenType.PLUS, self.parse_infix_expr)
        self.register_infix_fn(TokenType.SUB, self.parse_infix_expr)
        self.register_infix_fn(TokenType.MUL, self.parse_infix_expr)
        self.register_infix_fn(TokenType.DIV, self.parse_infix_expr)
        self.register_infix_fn(TokenType.EQ, self.parse_infix_expr)
        self.register_infix_fn(TokenType.NOT_EQ, self.parse_infix_expr)
        self.register_infix_fn(TokenType.LT, self.parse_infix_expr)
        self.register_infix_fn(TokenType.GT, self.parse_infix_expr)

    def next_token(self):
        self.cur_token = self.peek_token
        self.peek_token = self.lexer.next_token()

    def parse_program(self):
        self.sqss = StyleSheet()
        self.sqss.statements = []

        while self.peek_token.token_type is not TokenType.EOF:
            stmt = self.parse_stmt()
            if stmt is not None:
                self.sqss.statements.append(stmt)
            self.next_token()
        return self.sqss

    def parse_stmt(self):
        if self.cur_token.token_type == TokenType.IDENTIFIER or self.cur_token.token_type == TokenType.PROPERTY and self.peek_token.token_type == TokenType.ASSIGN:
            return self.parse_var_stmt()
        elif self.cur_token.token_type in [TokenType.CLASS_SELECTOR, TokenType.TYPE_SELECTOR]:
            return self.parse_ruleset()
        elif self.cur_token.token_type == TokenType.IMPORT:
            return self.parse_import()
        elif self.cur_token.token_type == TokenType.MIXIN:
            return self.parse_mixin()
        elif self.cur_token.token_type == TokenType.INCLUDE:
            return self.parse_include()
        elif self.cur_token.token_type == TokenType.EXTEND:
            return self.parse_extend()
        elif self.cur_token.token_type == TokenType.IF:
            return self.parse_if_stmt()
        else:
            return self.parse_expr_stmt()

    def parse_var_stmt(self):
        var_stmt = VarStatement()
        var_stmt.name = Identifier(self.cur_token, self.cur_token.literal)
        self.next_token()
        self.next_token()
        var_stmt.value = self.parse_expr_stmt()  # TODO 计算表达式

        return var_stmt

    def parse_prop_stmt(self):
        rule_stmt = Rule()
        rule_stmt.property = self.cur_token
        self.next_token()
        self.next_token()
        rule_stmt.value = self.parse_expr_stmt()  # TODO 计算表达式

        return rule_stmt

    def register_prefix_fn(self, token_type: TokenType, fn):
        self.prefix_parse_fns[token_type] = fn

    def register_infix_fn(self, token_type: TokenType, fn):
        self.infix_parse_fns[token_type] = fn

    def parse_expr_stmt(self):
        expr_stmt = ExpressionStatement()
        expr_stmt.expr = self.parse_expr(Priority.LOWEST)
        if self.peek_token.token_type == TokenType.SEMICOLON:
            self.next_token()
        return expr_stmt

    def parse_expr(self, precedence: int):

        prefix = self.prefix_parse_fns.get(self.cur_token.token_type)
        if not prefix:
            raise "prefix parse handler does not exists"
        left_expr = prefix()

        while self.peek_token != TokenType.SEMICOLON and precedence < self.peek_precedence():
            infix = self.infix_parse_fns.get(self.peek_token.token_type)
            if infix is None:
                return left_expr

            self.next_token()
            left_expr = infix(left_expr)

        return left_expr

    def parse_identifier(self):
        return Identifier(self.cur_token, self.cur_token.literal)

    def parse_color(self):
        return Color(self.cur_token)

    def parse_builtin(self):
        lexeme = self.cur_token
        self.next_token()
        args = self.parse_include_params()
        self.next_token()
        return Builtin(lexeme, args)

    def parse_integer_literal(self):
        lit = IntegerLiteral(self.cur_token)
        value = float(self.cur_token.literal)
        lit.value = value

        return lit

    def parse_boolean(self):
        return Boolean(self.cur_token, self.cur_token.token_type == TokenType.TRUE)

    def parse_prefix_expr(self):
        expr = PrefixExpression(self.cur_token, self.cur_token.literal)
        self.next_token()
        expr.right = self.parse_expr(Priority.PREFIX)

        return expr

    def peek_precedence(self):
        return precedences.get(self.peek_token.token_type, Priority.LOWEST)

    def cur_precedence(self):
        return precedences.get(self.cur_token.token_type, Priority.LOWEST)

    def parse_infix_expr(self, left: Expression):
        expr = InfixExpression(self.cur_token, self.cur_token.literal, left)

        precedence = self.cur_precedence()
        self.next_token()
        expr.right = self.parse_expr(precedence)

        return expr

    def parse_grouped_expr(self):
        self.next_token()

        expr = self.parse_expr(Priority.LOWEST)

        if self.peek_token.token_type != TokenType.RIGHT_PAREN:
            return None
        self.next_token()
        return expr

    def parse_import(self):
        self.next_token()
        with open(self.cur_token.literal) as f:
            lex = Lexer(f.read())
            p = Parser(lex)
            program = p.parse_program()
            self.sqss.statements.extend(program.statements)

    def parse_mixin(self):
        mixin_stmt = Mixin(self.cur_token)
        self.next_token()
        mixin_stmt.name = self.cur_token.literal

        if self.peek_token.token_type == TokenType.LEFT_PAREN:
            self.next_token()
            mixin_stmt.params = self.parse_mixin_params()
            if self.peek_token.token_type != TokenType.RIGHT_PAREN:
                return None
            self.next_token()

        mixin_stmt.body = Ruleset()
        mixin_stmt.body.selectors = []

        if self.peek_token.token_type != TokenType.LEFT_BRACE:
            return None
        self.next_token()
        rules = []
        child_ruleset_list = []
        self.next_token()
        while self.cur_token.token_type != TokenType.RIGHT_BRACE:
            if self.cur_token.token_type == TokenType.PROPERTY:
                rule = self.parse_rule()
                rules.append(rule)
            else:
                child_ruleset = self.parse_ruleset()
                child_ruleset_list.append(child_ruleset)
            self.next_token()
        # self.next_token()

        mixin_stmt.body.rules = rules
        mixin_stmt.body.child_rulesets = child_ruleset_list

        return mixin_stmt

    def parse_include(self):
        include_stmt = Include(self.cur_token)

        self.next_token()
        include_stmt.mixin_name = self.cur_token.literal

        if self.peek_token.token_type == TokenType.LEFT_PAREN:
            self.next_token()
            include_stmt.args = self.parse_include_params()

            if self.peek_token.token_type != TokenType.RIGHT_PAREN:
                return None
        self.next_token()

        return include_stmt

    def parse_extend(self):
        pass

    def parse_mixin_params(self):
        identifiers: list[Identifier] = []

        if self.peek_token.token_type == TokenType.RIGHT_PAREN:
            return identifiers

        self.next_token()

        ident = Identifier(self.cur_token, self.cur_token.literal)
        identifiers.append(ident)

        while self.peek_token.token_type == TokenType.COMMA:
            self.next_token()
            self.next_token()
            ident = Identifier(self.cur_token, self.cur_token.literal)
            identifiers.append(ident)

        if self.peek_token.token_type != TokenType.RIGHT_PAREN:
            return None

        return identifiers

    def parse_include_params(self):
        args: list[Identifier] = []

        if self.peek_token.token_type == TokenType.RIGHT_PAREN:
            return args

        self.next_token()

        arg = self.parse_expr(Priority.LOWEST)
        args.append(arg)

        while self.peek_token.token_type == TokenType.COMMA:
            self.next_token()
            self.next_token()
            arg = self.parse_expr(Priority.LOWEST)
            args.append(arg)

        if self.peek_token.token_type != TokenType.RIGHT_PAREN:
            return None

        return args

    def parse_selectors(self):
        selector_list: list[Selector] = []

        while self.peek_token.token_type != TokenType.LEFT_BRACE:
            tok = self.cur_token
            selector = Selector(tok)
            selector_list.append(selector)
            self.next_token()

        selector = Selector(self.cur_token)
        selector_list.append(selector)

        return selector_list

    def parse_ruleset(self):
        ruleset = Ruleset()
        ruleset.selectors = self.parse_selectors()

        if self.peek_token.token_type != TokenType.LEFT_BRACE:
            return None
        self.next_token()
        rules = []
        child_ruleset_list = []
        self.next_token()
        while self.cur_token.token_type != TokenType.RIGHT_BRACE:
            if self.cur_token.token_type == TokenType.PROPERTY:
                rule = self.parse_rule()
                rules.append(rule)
            elif self.cur_token.token_type == TokenType.INCLUDE:
                inc = self.parse_include()
                ruleset.includes.append(inc)
            else:
                child_ruleset = self.parse_ruleset()
                child_ruleset_list.append(child_ruleset)
            self.next_token()
        # self.next_token()

        ruleset.rules = rules
        ruleset.child_rulesets = child_ruleset_list

        return ruleset

    def parse_rule(self):
        rule = Rule()

        rule.property = self.cur_token
        self.next_token()
        self.next_token()
        rule.value = self.parse_expr(Priority.LOWEST)  # TODO 计算表达式

        self.next_token()
        return rule

    def parse_if_stmt(self):
        stmt = IfStatement(self.cur_token)
        self.next_token()

        stmt.condition = self.parse_expr(Priority.LOWEST)

        if self.peek_token.token_type != TokenType.LEFT_BRACE:
            return None
        self.next_token()

        stmt.consequence = self.parse_block_stmt()

        if self.peek_token.token_type == TokenType.ELSE:
            self.next_token()
            if self.peek_token.token_type != TokenType.LEFT_BRACE:
                return None
            self.next_token()
            stmt.alternative = self.parse_block_stmt()

        return stmt

    def parse_block_stmt(self):
        block = BlockStatement(self.cur_token)
        self.next_token()

        while self.cur_token.token_type != TokenType.RIGHT_BRACE:

            stmt = self.parse_stmt()
            if stmt:
                block.statements.append(stmt)
            self.next_token()  # 分号

        return block



import pqss


def test_mixin_parsing():
    tests = [('$a : 5; MyClass { width: $a; &:hover { width: $a + 3 * 5; &::indicator { width: $a + 3 * 5; } } } }', '$a', True)]
    for test in tests:
        qss = pqss.parse(test[0])
        print(qss)



from .ast import *
from .var import *
from .identifier import *
from .color import Color
from .builtin import Builtin
from .mixin import Mixin
from .include import Include
from .integer import IntegerLiteral
from .boolean import Boolean
from .selector import Selector
from .rule import Rule
from .ruleset import Ruleset
